
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program plaq_aidx()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "aflds.h"
#include "global.h"

static const int plns[6][2]={{0,1},{0,2},{0,3},{2,3},{3,1},{1,2}};


static void random_flds(int n,su3_alg_dble **wad)
{
   int i;

   for (i=0;i<n;i++)
   {
      if (cpr[0]==(NPROC0-1))
         gauss_dble((double*)(wad[i]),8*(4*VOLUME+7*(BNDRY/4)+3));
      else
         gauss_dble((double*)(wad[i]),8*(4*VOLUME+7*(BNDRY/4)));
   }
}


static void set_lks(su3_alg_dble *ad)
{
   int x0,x1,x2,x3,ix,mu;
   int N[4],bo[4],x[4],r;
   su3_alg_dble *X;

   N[0]=NPROC0*L0;
   N[1]=NPROC1*L1;
   N[2]=NPROC2*L2;
   N[3]=NPROC3*L3;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

               if (ix>=(VOLUME/2))
               {
                  x[0]=x0;
                  x[1]=x1;
                  x[2]=x2;
                  x[3]=x3;

                  X=ad+8*(ix-(VOLUME/2));

                  for (mu=0;mu<4;mu++)
                  {
                     (*X).c1=(double)(mu);
                     (*X).c2=(double)(bo[0]+x[0]);
                     (*X).c3=(double)(bo[1]+x[1]);
                     (*X).c4=(double)(bo[2]+x[2]);
                     (*X).c5=(double)(bo[3]+x[3]);
                     X+=1;

                     r=x[mu];
                     x[mu]-=1;

                     if ((bo[mu]==0)&&(x[mu]<0))
                        x[mu]+=N[mu];

                     (*X).c1=(double)(mu);
                     (*X).c2=(double)(bo[0]+x[0]);
                     (*X).c3=(double)(bo[1]+x[1]);
                     (*X).c4=(double)(bo[2]+x[2]);
                     (*X).c5=(double)(bo[3]+x[3]);
                     X+=1;
                     x[mu]=r;
                  }
               }
            }
         }
      }
   }
}


static void check_plaq_idx(su3_alg_dble *ad)
{
   int ix,iy,ie,n,mu,nu,r;
   int N[4],bo[4],x[4],ip[4];
   su3_alg_dble *X;

   N[0]=NPROC0*L0;
   N[1]=NPROC1*L1;
   N[2]=NPROC2*L2;
   N[3]=NPROC3*L3;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;
   ie=0;

   for (ix=0;ix<VOLUME;ix++)
   {
      iy=ix;
      x[3]=iy%L3;
      iy/=L3;
      x[2]=iy%L2;
      iy/=L2;
      x[1]=iy%L1;
      iy/=L1;
      x[0]=iy;

      iy=ipt[ix];

      for (n=0;n<6;n++)
      {
         mu=plns[n][0];
         nu=plns[n][1];
         plaq_aidx(n,iy,ip);

         X=ad+ip[0];
         ie|=((*X).c1!=(double)(mu));
         ie|=((*X).c2!=(double)(bo[0]+x[0]));
         ie|=((*X).c3!=(double)(bo[1]+x[1]));
         ie|=((*X).c4!=(double)(bo[2]+x[2]));
         ie|=((*X).c5!=(double)(bo[3]+x[3]));

         if ((n<3)&&(x[0]==(L0-1))&&(cpr[0]==(NPROC0-1)))
            ie|=(ip[1]!=(4*VOLUME+7*(BNDRY/4)+nu-1));
         else
         {
            r=x[mu];
            x[mu]+=1;

            if ((bo[mu]+x[mu])==N[mu])
               x[mu]-=N[mu];

            X=ad+ip[1];
            ie|=((*X).c1!=(double)(nu));
            ie|=((*X).c2!=(double)(bo[0]+x[0]));
            ie|=((*X).c3!=(double)(bo[1]+x[1]));
            ie|=((*X).c4!=(double)(bo[2]+x[2]));
            ie|=((*X).c5!=(double)(bo[3]+x[3]));
            x[mu]=r;
         }

         X=ad+ip[2];
         ie|=((*X).c1!=(double)(nu));
         ie|=((*X).c2!=(double)(bo[0]+x[0]));
         ie|=((*X).c3!=(double)(bo[1]+x[1]));
         ie|=((*X).c4!=(double)(bo[2]+x[2]));
         ie|=((*X).c5!=(double)(bo[3]+x[3]));

         r=x[nu];
         x[nu]+=1;

         if ((bo[nu]+x[nu])==N[nu])
            x[nu]-=N[nu];

         X=ad+ip[3];
         ie|=((*X).c1!=(double)(mu));
         ie|=((*X).c2!=(double)(bo[0]+x[0]));
         ie|=((*X).c3!=(double)(bo[1]+x[1]));
         ie|=((*X).c4!=(double)(bo[2]+x[2]));
         ie|=((*X).c5!=(double)(bo[3]+x[3]));
         x[nu]=r;
      }
   }

   error(ie!=0,1,"check_plaq_idx [check3.c]",
         "Incorrect offsets or incorrectly copied field variables");
}


int main(int argc,char *argv[])
{
   int my_rank;
   su3_alg_dble **wad;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);

      printf("\n");
      printf("Check of the program plaq_aidx()\n");
      printf("--------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
   }

   start_ranlux(0,12345);
   geometry();
   alloc_wad(2);
   wad=reserve_wad(2);
   random_flds(2,wad);

   set_lks(wad[0]);
   copy_bnd_ad(0,wad[0]);
   check_plaq_idx(wad[0]);

   set_lks(wad[1]);
   copy_bnd_ad(0,wad[1]);
   check_plaq_idx(wad[1]);

   release_wad();

   if (my_rank==0)
   {
      printf("No errors discovered\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
