
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2015 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Translation invariance of ghost_action().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "cflds.h"
#include "global.h"

#define NMAX 4

static double S0[NMAX+1],S1[NMAX+1],dS[NMAX+1];
static su3_alg_dble *Cw[NMAX+1];


static void new_fld(int n,su3_alg_dble **Cd)
{
   int i;

   for (i=0;i<=n;i++)
   {
      gauss_dble((double*)(Cd[i]),NSPIN);
      set_cdbc(0,Cd[i]);
   }
}


static void reorder_cd(int n,su3_alg_dble **Cd,su3_alg_dble *cd)
{
   int m,i,ix;

   m=n+1;

   for (i=0;i<m;i++)
      Cw[i]=Cd[i];

   for (ix=0;ix<VOLUME;ix++)
      for (i=0;i<m;i++)
         cd[ix*m+i]=Cw[i][ix];

   error(check_cdbc(n,cd)!=1,1,"reorder_cd [check4.c]",
         "Reordered field does not satisfy the proper "
         "boundary conditions");
}


static void random_vec(int *svec)
{
   int mu,bs[4];
   double r[4];

   bs[0]=NPROC0*L0;
   bs[1]=NPROC1*L1;
   bs[2]=NPROC2*L2;
   bs[3]=NPROC3*L3;
   svec[0]=0;
   svec[1]=0;
   svec[2]=0;
   svec[3]=0;

   while ((abs(svec[1])+abs(svec[2])+abs(svec[3]))==0)
   {
      ranlxd(r,4);

      for (mu=1;mu<4;mu++)
      {
         svec[mu]=(int)((double)(bs[mu])*r[mu]);
         if (svec[mu]>(bs[mu]/2))
            svec[mu]-=bs[mu];
      }
   }

   MPI_Bcast(svec,4,MPI_INT,0,MPI_COMM_WORLD);
}



int main(int argc,char *argv[])
{
   int my_rank,bc,i,n,s[4];
   su3_alg_dble **wcd,*cd;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check4.log","w",stdout);

      printf("\n");
      printf("Translation invariance of ghost_action()\n");
      printf("----------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check4.c]",
                    "Syntax: check4 [-bc <type>]");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   alloc_wcd(2*(NMAX+1));
   wcd=reserve_wcd(NMAX+1);
   cd=wcd[0];
   wcd=reserve_wcd(NMAX+1);

   if (my_rank==0)
   {
      printf("Translation invariance\n");
      printf("----------------------\n\n");
   }

   for (n=0;n<=NMAX;n++)
   {
      new_fld(n,wcd);
      reorder_cd(n,wcd,cd);
      ghost_action(1,n,cd,S0);
      random_vec(s);

      for (i=0;i<=n;i++)
         shift_cd(s,wcd[i]);

      reorder_cd(n,wcd,cd);
      ghost_action(1,n,cd,S1);

      for (i=0;i<=n;i++)
         dS[i]=S0[i]-S1[i];

      if (my_rank==0)
      {
         printf("n = %d, s=(% d,% d,% d,% d):\n",n,s[0],s[1],s[2],s[3]);

         for (i=0;i<=n;i++)
            printf("i = %d, action = %+.6e, relative deviation = %.1e\n",
                   i,S0[i],fabs(dS[i]/S0[i]));

         printf("\n");
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
