
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Gauge and translation invariance of action0().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "aflds.h"
#include "cflds.h"
#include "uflds.h"
#include "linalg.h"
#include "mdint.h"
#include "forces.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define NMAX 4

static int opv[(int)(PM_TYPES)];
static double *pv[(int)(PM_TYPES)];
static double S0[NMAX+1],S1[NMAX+1],dS[NMAX+1];
static su3_alg_dble *Aw;
static su3_dble *Uw;


static void default_pv(int order)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));
   error(p==NULL,1,"default_pv [check3.c]",
         "Unable to allocated parameter array");

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=0;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;

   for (i=2;i<=order;i+=2)
   {
      opv[(int)(CG)]=i;
      pv[(int)(CG)][i]=0.126-0.378*(double)(i*i/(order*order));
   }

   for (i=2;i<=order;i+=2)
   {
      opv[(int)(CGP)]=i;
      pv[(int)(CGP)][i]=0.0523+0.117*(double)(i*i/(order*order));
   }

   set_nspt_parms(KSPT,order,opv,pv);
}


static void alloc_arrays(void)
{
   int m;

   m=NMAX+1;
   Aw=amalloc(m*sizeof(*Aw),4);
   Uw=amalloc(m*sizeof(*Uw),4);

   error((Aw==NULL)||(Uw==NULL),1,"alloc_arrays [check3.c]",
         "Unable to allocate auxiliary arrays");
}


static void new_fld(int n,su3_alg_dble **ad)
{
   int i;

   for (i=0;i<=n;i++)
   {
      if (cpr[0]==(NPROC0-1))
         gauss_dble((double*)(ad[i]),8*(4*VOLUME+7*(BNDRY/4)+3));
      else
         gauss_dble((double*)(ad[i]),8*(4*VOLUME+7*(BNDRY/4)));

      set_adbc(0,ad[i]);
   }
}


void random_gauge_transform(int n,su3_dble *ud)
{
   int bc,m,ix,iy,t,ifc;
   double c;
   su3_alg_dble **wcd,*om;
   su3_dble **wgd,*gd;

   bc=bc_type();
   m=n+1;

   wcd=reserve_wcd(m);
   wgd=reserve_wgd(m);
   om=wcd[0];
   gd=wgd[0];

   gauss_dble((double*)(om),8*NSPIN*m);
   set_cdbc(n,om);
   cpcd_int_bnd(n,om);
   c=0.1;

   for (ix=0;ix<NSPIN;ix++)
   {
      mulr_alg(m,c,om,Aw);
      pexp_su3alg(m,Aw,gd);

      om+=m;
      gd+=m;
   }

   gd=wgd[0];

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            if (ifc&0x1)
            {
               iy=idn[ix][ifc/2];

               if ((t==0)&&(ifc==1))
                  cm3x3_varmul_poly(m,gd+iy*m,ud,ud);
               else
               {
                  cm3x3_dagger(m,gd+ix*m,Uw);
                  cm3x3_varmul_poly(m,ud,Uw,Uw);
                  cm3x3_varmul_poly(m,gd+iy*m,Uw,ud);
               }
            }
            else
            {
               iy=iup[ix][ifc/2];

               if ((t==(N0-1))&&(ifc==0))
                  cm3x3_varmul_poly(m,gd+ix*m,ud,ud);
               else
               {
                  cm3x3_dagger(m,gd+iy*m,Uw);
                  cm3x3_varmul_poly(m,ud,Uw,Uw);
                  cm3x3_varmul_poly(m,gd+ix*m,Uw,ud);
               }
            }
         }

         ud+=m;
      }
   }

   release_wgd();
   release_wcd();
}


static void random_vec(int *svec)
{
   int mu,bs[4];
   double r[4];

   bs[0]=NPROC0*L0;
   bs[1]=NPROC1*L1;
   bs[2]=NPROC2*L2;
   bs[3]=NPROC3*L3;
   svec[0]=0;
   svec[1]=0;
   svec[2]=0;
   svec[3]=0;

   while ((abs(svec[1])+abs(svec[2])+abs(svec[3]))==0)
   {
      ranlxd(r,4);

      for (mu=1;mu<4;mu++)
      {
         svec[mu]=(int)((double)(bs[mu])*r[mu]);
         if (svec[mu]>(bs[mu]/2))
            svec[mu]-=bs[mu];
      }
   }

   MPI_Bcast(svec,4,MPI_INT,0,MPI_COMM_WORLD);
}



int main(int argc,char *argv[])
{
   int my_rank,bc,n,i;
   int s[4];
   su3_alg_dble **wad;
   su3_dble **wud;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);

      printf("\n");
      printf("Gauge and translation invariance of action0()\n");
      printf("---------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check3.c]",
                    "Syntax: check3 [-bc <type>]");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   default_pv(NMAX);
   print_th_parms();
   print_nspt_parms((1<<(int)(LAMBDA))^0xffff);
   start_ranlux(0,12345);
   geometry();
   alloc_arrays();
   alloc_wad(NMAX+1);
   alloc_wud(NMAX+1);
   alloc_wcd(NMAX+1);
   alloc_wgd(NMAX+1);
   wad=reserve_wad(NMAX+1);
   wud=reserve_wud(NMAX+1);

   if (my_rank==0)
   {
      printf("Gauge invariance\n");
      printf("----------------\n\n");
   }

   for (n=0;n<=NMAX;n++)
   {
      new_fld(n,wad);
      a2uflds(n,wad,wud[0]);
      action0(1,n,wud[0],S0);
      random_gauge_transform(n,wud[0]);
      action0(1,n,wud[0],S1);

      for (i=0;i<=n;i++)
         dS[i]=S1[i]-S0[i];

      if (my_rank==0)
      {
         printf("n = %d:\n",n);

         for (i=0;i<=n;i++)
            printf("i = %d, relative deviation = %.1e\n",
                   i,fabs(1.0-S1[i]/S0[i]));

         printf("\n");
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      printf("Translation invariance\n");
      printf("----------------------\n\n");
   }

   for (n=0;n<=NMAX;n++)
   {
      new_fld(n,wad);
      a2uflds(n,wad,wud[0]);
      action0(1,n,wud[0],S0);
      random_vec(s);

      for (i=0;i<=n;i++)
         shift_ad(s,wad[i]);

      a2uflds(n,wad,wud[0]);
      action0(1,n,wud[0],S1);

      for (i=0;i<=n;i++)
         dS[i]=S1[i]-S0[i];

      if (my_rank==0)
      {
         printf("n = %d, s=(% d,% d,% d,% d):\n",n,s[0],s[1],s[2],s[3]);

         for (i=0;i<=n;i++)
            printf("i = %d, relative deviation = %.1e\n",
                   i,fabs(1.0-S1[i]/S0[i]));

         printf("\n");
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
