
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2015 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs set_cdbc() and check_cdbc().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "global.h"

#define NMAX 4

static su3_alg_dble ct0[NMAX];


static void new_flds(int n,su3_alg_dble *cd)
{
   gauss_dble((double*)(cd),8*NSPIN*(n+1));
}


static int is_equal(su3_alg_dble *cd1,su3_alg_dble *cd2)
{
   int ie;

   ie=1;
   ie&=((*cd1).c1==(*cd2).c1);
   ie&=((*cd1).c2==(*cd2).c2);
   ie&=((*cd1).c3==(*cd2).c3);
   ie&=((*cd1).c4==(*cd2).c4);
   ie&=((*cd1).c5==(*cd2).c5);
   ie&=((*cd1).c6==(*cd2).c6);
   ie&=((*cd1).c7==(*cd2).c7);
   ie&=((*cd1).c8==(*cd2).c8);

   return ie;
}


static void set_ct0(int n,su3_alg_dble *cd)
{
   int ip,iy,i,m;
   int cp[4];

   cp[0]=0;
   cp[1]=0;
   cp[2]=0;
   cp[3]=0;

   ip=ipr_global(cp);
   iy=ipt[0];
   m=n+1;

   for (i=0;i<m;i++)
      ct0[i]=cd[iy*m+i];

   MPI_Bcast(ct0,8*m,MPI_DOUBLE,ip,MPI_COMM_WORLD);
}


static int check_cdbnd(int n,su3_alg_dble *cd)
{
   int ix,i,m,ie;

   m=n+1;
   ie=1;

   if (bc_type()==1)
   {
      set_ct0(n,cd);

      for (ix=0;ix<VOLUME;ix++)
      {
         if (global_time(ix)==0)
         {
            for (i=0;i<m;i++)
               ie&=is_equal(ct0+i,cd+ix*m+i);
         }
      }
   }

   return ie;
}


static int cmp_cd(int n,su3_alg_dble *cd1,su3_alg_dble *cd2)
{
   int ix,i,m,ie;

   m=n+1;
   ie=1;

   if (bc_type()==1)
   {
      for (ix=0;ix<VOLUME;ix++)
      {
         if (global_time(ix)!=0)
         {
            for (i=0;i<m;i++)
               ie&=is_equal(cd1+ix*m+i,cd2+ix*m+i);
         }
      }
   }
   else
   {
      for (ix=0;ix<VOLUME;ix++)
      {
         for (i=0;i<m;i++)
            ie&=is_equal(cd1+ix*m+i,cd2+ix*m+i);
      }
   }

   for (ix=VOLUME;ix<NSPIN;ix++)
   {
      for (i=0;i<m;i++)
         ie&=is_equal(cd1+ix*m+i,cd2+ix*m+i);
   }

   return ie;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,ie;
   su3_alg_dble **wcd;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);
      printf("\n");
      printf("Check of set_cdbc() and check_cdbc()\n");
      printf("------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check3.c]",
                    "Syntax: check3 [-bc <type>]");
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   alloc_wcd(3*NMAX);

   for (n=0;n<NMAX;n++)
   {
      wcd=reserve_wcd(2*(n+1));
      new_flds(n,wcd[0]);
      new_flds(n,wcd[n+1]);

      assign_alg2alg(NSPIN*(n+1),wcd[0],wcd[n+1]);
      set_cdbc(n,wcd[0]);

      ie=check_cdbnd(n,wcd[0]);
      error(ie!=1,1,"main [check3.c]",
            "Boundary values are not correctly set");

      ie=check_cdbc(n,wcd[0]);
      error(ie!=1,1,"main [check3.c]",
            "check_cdbc() is not correct ");

      if (bc==1)
      {
         ie=check_cdbc(n,wcd[n+1]);
         error(ie!=0,2,"main [check3.c]",
               "check_cdbc() is not correct ");
      }

      ie=cmp_cd(n,wcd[0],wcd[n+1]);
      error(ie!=1,1,"main [check3.c]",
            "set_cdbc() changed field where it should not");

      release_wcd();
   }

   if (my_rank==0)
   {
      printf("No errors detected\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
