
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2015 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs set_udbc() and check_udbc().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "aflds.h"
#include "su3fcts.h"
#include "lattice.h"
#include "global.h"

#define NMAX 4

static void new_flds(int n,su3_dble *ud)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ud),18*(4*VOLUME+7*(BNDRY/4)+3)*(n+1));
   else
      gauss_dble((double*)(ud),18*(4*VOLUME+7*(BNDRY/4))*(n+1));
}


static int is_equal(su3_dble *ud1,su3_dble *ud2)
{
   int ie;

   ie=1;
   ie&=((*ud1).c11.re==(*ud2).c11.re);
   ie&=((*ud1).c11.im==(*ud2).c11.im);
   ie&=((*ud1).c12.re==(*ud2).c12.re);
   ie&=((*ud1).c12.im==(*ud2).c12.im);
   ie&=((*ud1).c13.re==(*ud2).c13.re);
   ie&=((*ud1).c13.im==(*ud2).c13.im);

   ie&=((*ud1).c21.re==(*ud2).c21.re);
   ie&=((*ud1).c21.im==(*ud2).c21.im);
   ie&=((*ud1).c22.re==(*ud2).c22.re);
   ie&=((*ud1).c22.im==(*ud2).c22.im);
   ie&=((*ud1).c23.re==(*ud2).c23.re);
   ie&=((*ud1).c23.im==(*ud2).c23.im);

   ie&=((*ud1).c31.re==(*ud2).c31.re);
   ie&=((*ud1).c31.im==(*ud2).c31.im);
   ie&=((*ud1).c32.re==(*ud2).c32.re);
   ie&=((*ud1).c32.im==(*ud2).c32.im);
   ie&=((*ud1).c33.re==(*ud2).c33.re);
   ie&=((*ud1).c33.im==(*ud2).c33.im);

   return ie;
}


static int is_zero(su3_dble *ud)
{
   int ie;

   ie=1;
   ie&=((*ud).c11.re==0.0);
   ie&=((*ud).c11.im==0.0);
   ie&=((*ud).c12.re==0.0);
   ie&=((*ud).c12.im==0.0);
   ie&=((*ud).c13.re==0.0);
   ie&=((*ud).c13.im==0.0);

   ie&=((*ud).c21.re==0.0);
   ie&=((*ud).c21.im==0.0);
   ie&=((*ud).c22.re==0.0);
   ie&=((*ud).c22.im==0.0);
   ie&=((*ud).c23.re==0.0);
   ie&=((*ud).c23.im==0.0);

   ie&=((*ud).c31.re==0.0);
   ie&=((*ud).c31.im==0.0);
   ie&=((*ud).c32.re==0.0);
   ie&=((*ud).c32.im==0.0);
   ie&=((*ud).c33.re==0.0);
   ie&=((*ud).c33.im==0.0);

   return ie;
}


static int cmp_active(int n,su3_dble *ud1,su3_dble *ud2)
{
   int bc,ix,t,ifc,k,m,ie;

   bc=bc_type();
   m=n+1;
   ie=1;

   for (ix=(VOLUME/2);ix<(VOLUME);ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((ifc<2)||(t>0)||(bc==2))
         {
            for (k=0;k<m;k++)
               ie&=is_equal(ud1+k,ud2+k);
         }

         ud1+=m;
         ud2+=m;
      }
   }

   return ie;
}


static int check_bval(int n,su3_dble *ud)
{
   int bc,ix,t,ifc,k,m,ie;

   bc=bc_type();
   m=n+1;
   ie=1;

   for (ix=(VOLUME/2);ix<(VOLUME);ix++)
   {
      t=global_time(ix);

      if ((bc==1)&&(t==0))
      {
         for (ifc=0;ifc<8;ifc++)
         {
            if (ifc>1)
            {
               for (k=0;k<m;k++)
                  ie&=is_zero(ud+k);
            }

            ud+=m;
         }
      }
      else
         ud+=(8*m);
   }

   if (cpr[0]==(NPROC0-1))
   {
      ud+=(7*m*(BNDRY/4));

      for (k=0;k<(3*m);k++)
         ie&=is_zero(ud+k);
   }

   return ie;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,ie;
   su3_dble **wud;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check4.log","w",stdout);
      printf("\n");
      printf("Check of set_udbc() and check_udbc()\n");
      printf("------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check4.c]",
                    "Syntax: check4 [-bc <type>]");
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   alloc_wud(2*NMAX);

   for (n=0;n<NMAX;n++)
   {
      wud=reserve_wud(2*(n+1));
      new_flds(n,wud[0]);
      new_flds(n,wud[n+1]);

      cm3x3_assign((4*VOLUME+7*(BNDRY/4))*(n+1),wud[0],wud[n+1]);
      set_udbc(n,wud[0]);

      ie=check_udbc(n,wud[0]);
      error(ie!=1,1,"main [check4.c]",
            "check_udbc() is incorrect");

      ie=check_udbc(n,wud[n+1]);
      error(ie!=0,2,"main [check4.c]",
            "check_udbc() is incorrect");

      ie=cmp_active(n,wud[0],wud[n+1]);
      error(ie!=1,1,"main [check4.c]",
            "Active link variables are modified by set_udbc()");

      ie=check_bval(n,wud[0]);
      error(ie!=1,1,"main [check4.c]",
            "Boundary values are not properly set by set_udbc()");

      release_wud();
   }

   if (my_rank==0)
   {
      printf("No errors detected\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
