
/*******************************************************************************
*
* File check6.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs cm3x3_mul_poly() and cm3x3_inv_poly().
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "random.h"
#include "su3fcts.h"

#define NTEST 100
#define SEED 376

static su3_dble wu[9] ALIGNED16;
static su3_dble wv[9] ALIGNED16;
static su3_dble ww[9] ALIGNED16;
static su3_dble wx[9] ALIGNED16;
static su3_dble one ALIGNED16;
static su3_dble zero ALIGNED16;


static double dev_uv(su3_dble *u,su3_dble *v)
{
   int i;
   double r[18],dev,dmax;

   r[ 0]=(*u).c11.re-(*v).c11.re;
   r[ 1]=(*u).c11.im-(*v).c11.im;
   r[ 2]=(*u).c12.re-(*v).c12.re;
   r[ 3]=(*u).c12.im-(*v).c12.im;
   r[ 4]=(*u).c13.re-(*v).c13.re;
   r[ 5]=(*u).c13.im-(*v).c13.im;

   r[ 6]=(*u).c21.re-(*v).c21.re;
   r[ 7]=(*u).c21.im-(*v).c21.im;
   r[ 8]=(*u).c22.re-(*v).c22.re;
   r[ 9]=(*u).c22.im-(*v).c22.im;
   r[10]=(*u).c23.re-(*v).c23.re;
   r[11]=(*u).c23.im-(*v).c23.im;

   r[12]=(*u).c31.re-(*v).c31.re;
   r[13]=(*u).c31.im-(*v).c31.im;
   r[14]=(*u).c32.re-(*v).c32.re;
   r[15]=(*u).c32.im-(*v).c32.im;
   r[16]=(*u).c33.re-(*v).c33.re;
   r[17]=(*u).c33.im-(*v).c33.im;

   dmax=0.0;

   for (i=0;i<18;i++)
   {
      dev=fabs(r[i]);
      if (dev>dmax)
         dmax=dev;
   }

   return dmax;
}


int main(void)
{
   int n,i,j;
   double r,d,dmax;

   printf("\n");
   printf("Check of the programs cm3x3_mul_poly() and cm3x3_inv_poly()\n");
   printf("-----------------------------------------------------------\n\n");

#if (defined AVX)
   printf("Using AVX instructions\n");
#elif (defined x64)
   printf("Using SSE3 instructions and up to 16 xmm registers\n");
#endif

   printf("Test performed on %d random matrices\n\n",NTEST);

   rlxd_init(1,SEED);
   cm3x3_unity(1,&one);
   cm3x3_zero(1,&zero);
   dmax=0.0;

   for (i=0;i<NTEST;i++)
   {
      for (n=1;n<=8;n++)
      {
         cm3x3_unity(1,wu);
         cm3x3_unity(1,wv);
         gauss_dble((double*)(wu+1),n*18);
         gauss_dble((double*)(wv+1),n*18);
         gauss_dble((double*)(ww),(n+1)*18);
         r=0.1;

         for (j=1;j<=n;j++)
            cm3x3_mulr(&r,wu+j,wu+j);

         cm3x3_inv_poly(n,wu+1,wv+1);
         cm3x3_mul_poly(n,wu,wv,ww);

         for (j=0;j<=n;j++)
         {
            if (j==0)
               d=dev_uv(ww+j,&one);
            else
               d=dev_uv(ww+j,&zero);

            if (d>dmax)
               dmax=d;
         }

         gauss_dble((double*)(wu),(n+1)*18);
         gauss_dble((double*)(wv),(n+1)*18);
         gauss_dble((double*)(ww),(n+1)*18);
         r=0.1;

         for (j=0;j<=n;j++)
         {
            cm3x3_mulr(&r,wu+j,wu+j);
            cm3x3_mulr(&r,wv+j,wv+j);
         }

         cm3x3_mul_poly(n,wu,wv,ww);
         cm3x3_mul_poly(n,wu,wv,wv);

         for (j=0;j<=n;j++)
         {
            d=dev_uv(ww+j,wv+j);

            if (d>dmax)
               dmax=d;
         }

         cm3x3_unity(1,wu);
         cm3x3_unity(1,wv);
         gauss_dble((double*)(wu+1),n*18);
         gauss_dble((double*)(wv+1),n*18);
         gauss_dble((double*)(ww),(n+1)*18);
         r=0.1;

         for (j=1;j<=n;j++)
         {
            cm3x3_mulr(&r,wu+j,wu+j);
            cm3x3_mulr(&r,wv+j,wv+j);
         }

         cm3x3_mul_poly(n,wu,wv,ww);
         cm3x3_varmul_poly(n,wu+1,wv+1,wx+1);

         for (j=1;j<=n;j++)
         {
            d=dev_uv(ww+j,wx+j);

            if (d>dmax)
               dmax=d;
         }

         cm3x3_varmul_poly(n,wu+1,wv+1,wv+1);

         for (j=1;j<=n;j++)
         {
            d=dev_uv(wv+j,wx+j);

            if (d>dmax)
               dmax=d;
         }
      }
   }

   printf("Maximal deviation = %.1e\n\n",dmax);
   exit(0);
}
