
/*******************************************************************************
*
* File check8.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs u3alg_zero(),..,u3alg_to_su3alg().
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "random.h"
#include "su3fcts.h"

#define NTEST 1000
#define SEED 376

static u3_alg_dble ad0={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
static u3_alg_dble wa[4];
static su3_alg_dble wsa[1] ALIGNED16;
static su3_dble wu[4] ALIGNED16;


static double dev_ab(u3_alg_dble *a,u3_alg_dble *b)
{
   int i;
   double r[9],dev,dmax;

   r[0]=(*a).c1-(*b).c1;
   r[1]=(*a).c2-(*b).c2;
   r[2]=(*a).c3-(*b).c3;
   r[3]=(*a).c4-(*b).c4;
   r[4]=(*a).c5-(*b).c5;
   r[5]=(*a).c6-(*b).c6;
   r[6]=(*a).c7-(*b).c7;
   r[7]=(*a).c8-(*b).c8;
   r[8]=(*a).c9-(*b).c9;

   dmax=0.0;

   for (i=0;i<9;i++)
   {
      dev=fabs(r[i]);
      if (dev>dmax)
         dmax=dev;
   }

   return dmax;
}


static double dev_au(u3_alg_dble *a,su3_dble *u)
{
   int i;
   double r[18],dev,dmax;

   r[ 0]=(*a).c1-(*u).c11.im;
   r[ 1]=(*u).c11.re;
   r[ 2]=(*a).c2-(*u).c22.im;
   r[ 3]=(*u).c22.re;
   r[ 4]=(*a).c3-(*u).c33.im;
   r[ 5]=(*u).c33.re;

   r[ 6]=(*a).c4-(*u).c12.re;
   r[ 7]=(*a).c4+(*u).c21.re;
   r[ 8]=(*a).c5-(*u).c12.im;
   r[ 9]=(*a).c5-(*u).c21.im;

   r[10]=(*a).c6-(*u).c13.re;
   r[11]=(*a).c6+(*u).c31.re;
   r[12]=(*a).c7-(*u).c13.im;
   r[13]=(*a).c7-(*u).c31.im;

   r[14]=(*a).c8-(*u).c23.re;
   r[15]=(*a).c8+(*u).c32.re;
   r[16]=(*a).c9-(*u).c23.im;
   r[17]=(*a).c9-(*u).c32.im;

   dmax=0.0;

   for (i=0;i<18;i++)
   {
      dev=fabs(r[i]);
      if (dev>dmax)
         dmax=dev;
   }

   return dmax;
}


static double dev_uv(su3_dble *u,su3_dble *v)
{
   int i;
   double r[18],dev,dmax;

   r[ 0]=(*u).c11.re-(*v).c11.re;
   r[ 1]=(*u).c11.im-(*v).c11.im;
   r[ 2]=(*u).c12.re-(*v).c12.re;
   r[ 3]=(*u).c12.im-(*v).c12.im;
   r[ 4]=(*u).c13.re-(*v).c13.re;
   r[ 5]=(*u).c13.im-(*v).c13.im;

   r[ 6]=(*u).c21.re-(*v).c21.re;
   r[ 7]=(*u).c21.im-(*v).c21.im;
   r[ 8]=(*u).c22.re-(*v).c22.re;
   r[ 9]=(*u).c22.im-(*v).c22.im;
   r[10]=(*u).c23.re-(*v).c23.re;
   r[11]=(*u).c23.im-(*v).c23.im;

   r[12]=(*u).c31.re-(*v).c31.re;
   r[13]=(*u).c31.im-(*v).c31.im;
   r[14]=(*u).c32.re-(*v).c32.re;
   r[15]=(*u).c32.im-(*v).c32.im;
   r[16]=(*u).c33.re-(*v).c33.re;
   r[17]=(*u).c33.im-(*v).c33.im;

   dmax=0.0;

   for (i=0;i<18;i++)
   {
      dev=fabs(r[i]);
      if (dev>dmax)
         dmax=dev;
   }

   return dmax;
}


int main(void)
{
   int i;
   double r,dev,dmax[9];

   printf("\n");
   printf("Check of the programs u3alg_zero(),..,u3alg_to_su3alg()\n");
   printf("-------------------------------------------------------\n\n");

#if (defined AVX)
   printf("Using AVX instructions\n");
#elif (defined x64)
   printf("Using SSE3 instructions and up to 16 xmm registers\n");
#endif

   printf("Test performed on %d random matrices\n\n",NTEST);

   rlxd_init(1,SEED);

   for (i=0;i<8;i++)
      dmax[i]=0.0;

   for (i=0;i<NTEST;i++)
   {
      gauss_dble((double*)(wa),4*9);
      gauss_dble((double*)(wu),4*18);

      u3alg_zero(wa);
      dev=dev_ab(&ad0,wa);
      if (dev>dmax[0])
         dmax[0]=dev;

      gauss_dble((double*)(wa),9);
      gauss_dble((double*)(wa+1),9);
      u3alg_to_su3(wa,wu);
      u3alg_to_su3(wa+1,wu+1);
      u3alg_add(wa,wa+1);
      cm3x3_add(wu,wu+1);
      dev=dev_au(wa+1,wu+1);
      if (dev>dmax[1])
         dmax[1]=dev;

      gauss_dble((double*)(wa),9);
      gauss_dble((double*)(wa+1),9);
      u3alg_to_su3(wa,wu);
      u3alg_to_su3(wa+1,wu+1);
      u3alg_sub(wa,wa+1);
      cm3x3_sub(wu,wu+1);
      dev=dev_au(wa+1,wu+1);
      if (dev>dmax[2])
         dmax[2]=dev;

      gauss_dble((double*)(wa),9);
      r=2.9782;
      u3alg_mulr(&r,wa,wa+1);
      u3alg_to_su3(wa,wu);
      cm3x3_mulr(&r,wu,wu);
      dev=dev_au(wa+1,wu);
      if (dev>dmax[3])
         dmax[3]=dev;
      u3alg_mulr(&r,wa,wa);
      dev=dev_au(wa+1,wu);
      if (dev>dmax[3])
         dmax[3]=dev;

      gauss_dble((double*)(wa),9);
      gauss_dble((double*)(wa+1),9);
      u3alg_to_su3(wa,wu);
      u3alg_to_su3(wa+1,wu+1);
      r=-0.78632;
      u3alg_mulr_add(&r,wa,wa+1);
      cm3x3_mulr_add(&r,wu,wu+1);
      dev=dev_au(wa+1,wu+1);
      if (dev>dmax[4])
         dmax[4]=dev;

      gauss_dble((double*)(wa),9);
      u3alg_to_su3(wa,wu);
      dev=dev_au(wa,wu);
      if (dev>dmax[5])
         dmax[5]=dev;

      gauss_dble((double*)(wu),18);
      su3_to_u3alg(wu,wa);
      cm3x3_dagger(1,wu,wu+1);
      cm3x3_sub(wu+1,wu);
      r=0.5;
      cm3x3_mulr(&r,wu,wu);
      dev=dev_au(wa,wu);
      if (dev>dmax[6])
         dmax[6]=dev;

      gauss_dble((double*)(wa),9);
      u3alg_to_su3alg(wa,wsa);
      u3alg_to_su3(wa,wu);
      su3alg_to_su3(wsa,wu+1);
      r=((*wu).c11.im+(*wu).c22.im+(*wu).c33.im)/3.0;
      (*wu).c11.im-=r;
      (*wu).c22.im-=r;
      (*wu).c33.im-=r;
      dev=dev_uv(wu,wu+1);
      if (dev>dmax[7])
         dmax[7]=dev;
   }

   printf("Maximal deviation of u3alg_zero()          = %.1e\n",dmax[0]);
   printf("Maximal deviation of u3alg_add()           = %.1e\n",dmax[1]);
   printf("Maximal deviation of u3alg_sub()           = %.1e\n",dmax[2]);
   printf("Maximal deviation of u3alg_mulr()          = %.1e\n",dmax[3]);
   printf("Maximal deviation of u3alg_mulr_add()      = %.1e\n",dmax[4]);
   printf("Maximal deviation of u3alg_to_su3()        = %.1e\n",dmax[5]);
   printf("Maximal deviation of su3_to_u3alg()        = %.1e\n",dmax[6]);
   printf("Maximal deviation of u3alg_to_su3alg()     = %.1e\n\n",dmax[7]);

   for (i=1;i<9;i++)
   {
      if (dmax[i]>dmax[0])
         dmax[0]=dmax[i];
   }

   printf("Maximal deviation (all tests)            = %.1e\n\n",dmax[0]);
   exit(0);
}
